(defpackage #:renderer
  (:use #:cl)
  (:export #:init-renderer #:render)
  (:import-from #:sb-cga #:vec #:vec* #:matrix* #:translate #:rotate)
  (:import-from #:kit.math #:deg-to-rad))

(in-package #:renderer)

(defparameter *vbo* 0)
(defparameter *vao* 0)
(defparameter *ebo* 0)
(defparameter *shader-program* nil)
(defparameter *light-shader* nil)

(defparameter *diffuse-map* nil)
(defparameter *specular-map* nil)

(defparameter *light-vao* 0)
(defparameter *light-pos* (sb-cga:vec 1.2 1.0 2.0))

(defun init-renderer ()
  (setf *shader-program* (shader:load #p"shaders/shader.vert" #p"shaders/shader.frag"))
  (setf *light-shader* (shader:load #p"shaders/light.vert" #p"shaders/light.frag"))

  (setf *diffuse-map* (texture:load #p"textures/container2.png"))
  (setf *specular-map* (texture:load #p"textures/container2_specular.png"))

  (setf *vbo* (gl:gen-buffer))
  (setf *ebo* (gl:gen-buffer))
  (setf *vao* (gl:gen-vertex-array))
  
  (gl:enable :depth-test)
  (gl:bind-vertex-array *vao*)
  (gl:bind-buffer :array-buffer *vbo*)
  (gl:bind-buffer :element-array-buffer *ebo*)
 ;;                   positions         normals         htexture coords
  (let* ((vertices #(-0.5 -0.5 -0.5    +0.0 +0.0 -1.0    0.0 0.0
		     +0.5 -0.5 -0.5    +0.0 +0.0 -1.0    1.0 0.0
		     +0.5 +0.5 -0.5    +0.0 +0.0 -1.0    1.0 1.0
		     -0.5 +0.5 -0.5    +0.0 +0.0 -1.0    0.0 1.0
		     -0.5 -0.5 -0.5    +0.0 +0.0 -1.0    0.0 0.0
				       
		     -0.5 -0.5 +0.5    +0.0 +0.0 +1.0    0.0 0.0
		     +0.5 -0.5 +0.5    +0.0 +0.0 +1.0    1.0 0.0
		     +0.5 +0.5 +0.5    +0.0 +0.0 +1.0    1.0 1.0
		     -0.5 +0.5 +0.5    +0.0 +0.0 +1.0    0.0 1.0
		     -0.5 -0.5 +0.5    +0.0 +0.0 +1.0    0.0 0.0
				       
		     -0.5 +0.5 +0.5    -1.0 +0.0 +0.0    1.0 0.0
		     -0.5 +0.5 -0.5    -1.0 +0.0 +0.0    1.0 1.0
		     -0.5 -0.5 -0.5    -1.0 +0.0 +0.0    0.0 1.0
		     -0.5 -0.5 +0.5    -1.0 +0.0 +0.0    0.0 0.0
		     -0.5 +0.5 +0.5    -1.0 +0.0 +0.0    1.0 0.0
				       
		     +0.5 +0.5 +0.5    +1.0 +0.0 +0.0    1.0 0.0
		     +0.5 +0.5 -0.5    +1.0 +0.0 +0.0    1.0 1.0
		     +0.5 -0.5 -0.5    +1.0 +0.0 +0.0    0.0 1.0
		     +0.5 -0.5 +0.5    +1.0 +0.0 +0.0    0.0 0.0
		     +0.5 +0.5 +0.5    +1.0 +0.0 +0.0    1.0 0.0
				       
		     -0.5 -0.5 -0.5    +0.0 -1.0 +0.0    0.0 1.0
		     +0.5 -0.5 -0.5    +0.0 -1.0 +0.0    1.0 1.0
		     +0.5 -0.5 +0.5    +0.0 -1.0 +0.0    1.0 0.0
		     -0.5 -0.5 +0.5    +0.0 -1.0 +0.0    0.0 0.0
		     -0.5 -0.5 -0.5    +0.0 -1.0 +0.0    0.0 1.0
				       
		     -0.5 +0.5 -0.5    +0.0 +1.0 +0.0    0.0 1.0
		     +0.5 +0.5 -0.5    +0.0 +1.0 +0.0    1.0 1.0
		     +0.5 +0.5 +0.5    +0.0 +1.0 +0.0    1.0 0.0
		     -0.5 +0.5 +0.5    +0.0 +1.0 +0.0    0.0 0.0
		     -0.5 +0.5 -0.5    +0.0 +1.0 +0.0    0.0 1.0))
	 (arr (gl:alloc-gl-array :float (length vertices))))
    (dotimes (i (length vertices))
      (setf (gl:glaref arr i) (aref vertices i)))
    (gl:buffer-data :array-buffer :static-draw arr)
    (gl:free-gl-array arr))
  
  (let* ((indices #(0  1  2    2  3  4
		    5  6  7    7  8  9
		    10 11 12   12 13 14
		    15 16 17   17 18 19
		    20 21 22   22 23 24
		    25 26 27   27 28 29))
	 (arr (gl:alloc-gl-array :int (length indices))))
    (dotimes (i (length indices))
      (setf (gl:glaref arr i) (aref indices i)))
    (gl:buffer-data :element-array-buffer :static-draw arr)
    (gl:free-gl-array arr))
  
  (gl:vertex-attrib-pointer 0 3 :float nil
			    (* 8 (cffi:foreign-type-size :float))
			    (cffi:null-pointer))
  (gl:enable-vertex-attrib-array 0)
  
  (gl:vertex-attrib-pointer 1 3 :float nil
			    (* 8 (cffi:foreign-type-size :float))
			    (*  3 (cffi:foreign-type-size :float)))
  (gl:enable-vertex-attrib-array 1)
  
  (gl:vertex-attrib-pointer 2 2 :float nil
			    (* 8 (cffi:foreign-type-size :float))
			    (* 6 (cffi:foreign-type-size :float)))
  (gl:enable-vertex-attrib-array 2)

  (setf *light-vao* (gl:gen-vertex-array))
  (gl:bind-vertex-array *light-vao*)
  (gl:bind-buffer :array-buffer *vbo*)
  (gl:bind-buffer :element-array-buffer *ebo*)
  (gl:vertex-attrib-pointer 0 3 :float nil
			    (* 8 (cffi:foreign-type-size :float))
			    (cffi:null-pointer))
  (gl:enable-vertex-attrib-array 0))

(defun enable-wireframe ()
 (gl:polygon-mode :front-and-back :line))

(defun disable-wireframe ()
 (gl:polygon-mode :front-and-back :fill))

(let ((wireframe nil))
  (defun toggle-wireframe ()
    (if wireframe
	(disable-wireframe)
	(enable-wireframe))
    (setf wireframe (not wireframe))))

(defun render ()
  (gl:clear-color 0.1 0.1 0.1 1.0)
  (gl:clear :color-buffer :depth-buffer-bit)

  (shader:use *shader-program*)
  (shader:set-int *shader-program* "material.diffuse" 0)
  (shader:set-int *shader-program* "material.specular" 1)
  (shader:set-vec *shader-program*   "viewPos" camera:*position*)
  (shader:set-float *shader-program* "material.shininess" 32.0)

  (gl:active-texture :texture0)
  (texture:bind *diffuse-map*)

  (gl:active-texture :texture1)
  (texture:bind *specular-map*)

  ;; Directional light
  (shader:set-vec *shader-program* "dirLight.direction" (vec -0.2  -1.0  -0.3))
  (shader:set-vec *shader-program* "dirLight.ambient"   (vec  0.05  0.05  0.05))
  (shader:set-vec *shader-program* "dirLight.diffuse"   (vec  0.4   0.4   0.4))
  (shader:set-vec *shader-program* "dirLight.specular"  (vec  0.5   0.5   0.5))

  ;; Point lights
  (let ((point-light-positions (list (vec +0.7 +0.2 +2.0)
				     (vec +2.3 -3.3 -4.0)
				     (vec -4.0 +2.0 -12.0)
				     (vec +0.0 +0.0 -3.0))))
    (dotimes (i (length point-light-positions))
      (shader:set-vec *shader-program* (format nil "pointLights[~a].position" i)
		      (nth i point-light-positions))
      (shader:set-vec *shader-program* (format nil "pointLights[~a].ambient" i)
		      (vec 0.05 0.05 0.05))
      (shader:set-vec *shader-program* (format nil "pointLights[~a].diffuse" i)
		      (vec 0.8 0.8 0.8))
      (shader:set-vec *shader-program* (format nil "pointLights[~a].specular" i)
		      (vec 1.0 1.0 1.0))
      (shader:set-float *shader-program* (format nil "pointLights[~a].constant" i) 1.0)
      (shader:set-float *shader-program* (format nil "pointLights[~a].linear" i) 0.09)
      (shader:set-float *shader-program* (format nil "pointLights[~a].quadratic" i) 0.032))

    ;; Spot light
    (shader:set-vec *shader-program*   "spotLight.position"  camera:*position*)
    (shader:set-vec *shader-program*   "spotLight.direction" camera:*front*)
    (shader:set-vec *shader-program*   "spotLight.ambient"   (vec 0.0 0.0 0.0))
    (shader:set-vec *shader-program*   "spotLight.diffuse"   (vec 1.0 1.0 1.0))
    (shader:set-vec *shader-program*   "spotLight.specular"  (vec 1.0 1.0 1.0))
    (shader:set-float *shader-program* "spotLight.constant"  1.0)
    (shader:set-float *shader-program* "spotLight.linear"    0.09)
    (shader:set-float *shader-program* "spotLight.quadratic" 0.032)
    (shader:set-float *shader-program* "spotLight.cutOff"    (cos (deg-to-rad 12.5)))
    (shader:set-float *shader-program* "spotLight.outerCutOff" (cos (deg-to-rad 15.0)))
    
    ;; view/projection transformations
    (shader:set-mat4 *shader-program* "projection"
		     (kit.math:perspective-matrix (kit.math:deg-to-rad camera:*zoom*)
						  (apply #'/ (glfw:get-window-size))
						  0.1
						  100.0)
		     nil)
    (shader:set-mat4 *shader-program* "view" (camera:get-view-matrix) nil)
    
    (gl:bind-vertex-array *vao*)
    (let ((cube-positions (list (vec +0.0 +0.0 +0.0)
				(vec +2.0 +5.0 -15.0)
				(vec -1.5 -2.2 -2.5)
				(vec -3.8 -2.0 -12.3)
				(vec +2.4 -0.4 -3.5)
				(vec -1.7 +3.0 -7.5)
				(vec +1.3 -2.0 -2.5)
				(vec +1.5 +2.0 -2.5)
				(vec +1.5 +0.2 -1.5)
				(vec -1.3 +1.0 -1.5))))
      (dotimes (i (length cube-positions))
	(shader:set-mat4 *shader-program*
			 "model"
			 (matrix* (translate (nth i cube-positions))
				  (rotate (vec* (vec 1.0 0.3 0.5)
						(deg-to-rad (* 20.0 i)))))
			 nil)
	(gl:draw-elements :triangles
			  (gl:make-null-gl-array :unsigned-int)
			  :count 36)))
    
    (shader:use *light-shader*)
    (shader:set-mat4 *light-shader* "view" (camera:get-view-matrix) nil)
    (shader:set-mat4 *light-shader*
		     "projection"
		     (kit.math:perspective-matrix (kit.math:deg-to-rad camera:*zoom*)
						  (apply #'/ (glfw:get-window-size))
						  0.1
						  100.0)
		     nil)
    (gl:bind-vertex-array *light-vao*)
    (dotimes (i (length point-light-positions))
      (shader:set-float *light-shader* (format nil "pointLights[~a].constant" 1) 1.0)
      (shader:set-mat4 *light-shader*
		       "model"
		       (sb-cga:matrix* (sb-cga:translate (nth i point-light-positions))
				       (sb-cga:scale* 0.2 0.2 0.2))
		       nil)
      (gl:draw-elements :triangles
			(gl:make-null-gl-array :unsigned-int)
			:count 36))
    )
  
  (glfw:swap-buffers))
