(in-package #:business-tycoon)

(defparameter *delta-time* 0)
(defparameter *last-frame* 0)

(defun close-window ()
  (glfw:set-window-should-close))

(glfw:def-key-callback key-callback (window key scancode action mod-keys)
  (declare (ignore window scancode mod-keys))
  (when (and (eq key :escape) (eq action :press))
    (glfw:set-window-should-close))
  (when (and (eq key :space) (eq action :press))
    (toggle-wireframe)))

(glfw:def-mouse-button-callback mouse-button-callback (window button action mod-keys)
  (declare (ignore window button action mod-keys)))

(let ((last-x 400) (last-y 300))
  (glfw:def-cursor-pos-callback cursor-pos-callback (window x y)
    (declare (ignore window))
    (camera:process-mouse-movement (- x last-x) (- last-y y))
    (setf last-x x)
    (setf last-y y))

  (glfw:def-cursor-enter-callback cursor-enter-callback (window enterp)
    (declare (ignore window))
    (when enterp
      (let ((cursor-position (glfw:get-cursor-position)))
	(setf last-x (first cursor-position))
	(setf last-y (second cursor-position))))))

(glfw:def-scroll-callback scroll-callback (window x y)
  (declare (ignore window x))
  (camera:process-mouse-scroll y))

(glfw:def-window-size-callback update-viewport (window w h)
  (declare (ignore window))
  (gl:viewport 0 0 w h))

(defun open-window (&key (title "") (width 800) (height 600))
  ;; (tmt:with-body-in-main-thread ()
    (glfw:with-init-window (:title title
			    :width width
			    :height height
			    ;; :context-version-major 3
			    ;; :context-version-minor 3
			    ;; :opengl-profile :opengl-core-profile
			    )
      (setf %gl:*gl-get-proc-address* #'glfw:get-proc-address)
      (glfw:set-key-callback 'key-callback)
      (glfw:set-mouse-button-callback 'mouse-button-callback)
      (glfw:set-window-size-callback 'update-viewport)
      (glfw:set-cursor-enter-callback 'cursor-enter-callback)
      (glfw:set-cursor-position-callback 'cursor-pos-callback)
      (glfw:set-scroll-callback 'scroll-callback)
      (glfw:set-input-mode :cursor :disabled)
      (gl:viewport 0 0 width height)
      (camera:make-camera (sb-cga:vec 1.0 2.0 6.0))
      (renderer:init-renderer)
      (loop until (glfw:window-should-close-p)
	    do (let ((current-frame (glfw:get-time)))
		 (setf *delta-time* (- current-frame *last-frame*))
		 (setf *last-frame* current-frame)
		 (when (eq (glfw:get-key :w) :press) (camera:move-forward *delta-time*))
		 (when (eq (glfw:get-key :s) :press) (camera:move-backward *delta-time*))
		 (when (eq (glfw:get-key :a) :press) (camera:move-left *delta-time*))
		 (when (eq (glfw:get-key :d) :press) (camera:move-right *delta-time*))
		 (renderer:render)
		 (glfw:poll-events)))))
;; )
